/* Write out a file in VDF -- Virtual world Description Format */

/* 
 *  This implementation written by:
 *      Bernie Roehl (broehl@uwaterloo.ca)
 *      November, 1994
 *  Redistribute at will.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>

#include "vdf.h"

/* These macros check to see if a color or a vector is not [0,0,0] */

#define color_non_zero(c) ((c)->red != 0.0 || (c)->green != 0.0 || (c)->blue != 0.0)
#define vector_non_zero(v) ((v)->x != 0.0 || (v)->y != 0.0 || (v)->z != 0.0)

static FILE *outfile = NULL;

/**** error-handling routines ****/

static jmp_buf error_jmp;

static char emsg[500];

static int error_setup(void)
	{
	return setjmp(error_jmp);
	}

char *vdf_get_write_error(void)
	{
	return emsg;
	}

static void error_abort(char *str)
	{
	strcpy(emsg, str);
	longjmp(error_jmp, 1);
	}

/**** error-handling versions of various routines ****/

static void *emalloc(unsigned int nbytes)
	{
	void *ptr = vdf_malloc(nbytes);
	if (ptr) return ptr;
	error_abort("out of memory");
	return NULL;  /* should never get here */
	}

static char *estrdup(char *string)
	{
	char *ptr = emalloc(strlen(string)+1);
	strcpy(ptr, string);
	return ptr;
	}

/* output ("dump") routines */

/* write out the palette */

static void dump_palette(VDF_PALETTE *pal)
	{
	if (pal == NULL)
		return;
	if (pal->ncolors == 0 && pal->nhues == 0)
		return;
	fprintf(outfile, "Palette\n\t{\n");
	if (pal->ncolors)
		{
		int i;
		fprintf(outfile, "\tColor_table\n\t\t{\n\t\tCount { %ld }\n", (long) pal->ncolors);
		for (i = 0; i < pal->ncolors; ++i)
			{
			VDF_COLOR *color = &pal->colors[i];
			fprintf(outfile, "\t\tColor_entry { %f %f %f }\n", color->red, color->green, color->blue);
			}
		fprintf(outfile, "\t\t}\n");
		}
	if (pal->nhues)
		{
		int i;
		fprintf(outfile, "\tHue_table\n\t\t{\n\t\tCount { %ld }\n", (long) pal->nhues);
		for (i = 0; i < pal->nhues; ++i)
			fprintf(outfile, "\t\tHue_entry { %ld %ld }\n", (long) pal->hues[i].darkest, (long) pal->hues[i].lightest);
		fprintf(outfile, "\t\t}\n");
		}
	fprintf(outfile, "\t}\n\n");
	}

/* dump a list of Maps */

static void dump_maps(VDF_MAP *map)
	{
	while (map)
		{
		if (map->name && map->filename)
			fprintf(outfile, "Map { Name { \"%s\" } Filename { \"%s\" } }\n\n",
				map->name, map->filename);
		map = map->next;
		}
	}

static char *render_type_names[] =
	{
	"WIREFRAME", "UNLIT", "FLAT", "GOURAUD", "PHONG"
	};

/* dump a list of Materials */

static void dump_materials(VDF_MATERIAL *material)
	{
	while (material)
		{
		fprintf(outfile, "Material\n\t{\n");
		fprintf(outfile, "\tIdentifier { 0x%lX }\n", material->id);
		if (material->name)
			fprintf(outfile, "\tName { \"%s\" }\n", material->name);
		if (material->rendering_mode != VDF_RENDER_FLAT)
			fprintf(outfile, "\tRendering_mode { %s }\n", render_type_names[material->rendering_mode]);
		if (material->hue != -1)
			fprintf(outfile, "\tHue { %d }\n", material->hue);
		if (color_non_zero(&material->diffuse_color))
			fprintf(outfile, "\tDiffuse_color { %f %f %f }\n",
				material->diffuse_color.red, material->diffuse_color.green, material->diffuse_color.blue);
		if (color_non_zero(&material->ambient_color))
			fprintf(outfile, "\tAmbient_color { %f %f %f }\n",
				material->ambient_color.red, material->ambient_color.green, material->ambient_color.blue);
		if (color_non_zero(&material->transparency_color))
			fprintf(outfile, "\tTransparency_color { %f %f %f }\n",
				material->transparency_color.red, material->transparency_color.green, material->transparency_color.blue);
		if (color_non_zero(&material->specular_color))
			fprintf(outfile, "\tSpecular_color { %f %f %f }\n",
				material->specular_color.red, material->specular_color.green, material->specular_color.blue);
		if (material->albedo != 0.5)
			fprintf(outfile, "\tAlbedo { %f }\n", material->albedo);
		if (material->specular_exponent != 0.00)
			fprintf(outfile, "\tSpecular_exponent { %f }\n", material->specular_exponent);
		if (material->refractive_index != 0.00)
			fprintf(outfile, "\tRefractive_index { %f }\n", material->refractive_index);
		if (material->reflection_blur != 0.00)
			fprintf(outfile, "\tReflection_blur { %f }\n", material->reflection_blur);
		if (material->transparency_falloff != 0.00)
			fprintf(outfile, "\tTransparency_falloff { %f }\n", material->transparency_falloff);
		if (material->texture_map)
			fprintf(outfile, "\tTexture_map { Name { \"%s\" } }\n", material->texture_map);
		if (material->bump_map)
			fprintf(outfile, "\tBump_map { Name { \"%s\" } }\n", material->bump_map);
		if (material->opacity_map)
			fprintf(outfile, "\tOpacity_map { Name { \"%s\" } }\n", material->opacity_map);
		if (material->reflection_map)
			fprintf(outfile, "\tReflection_map { Name { \"%s\" } }\n", material->reflection_map);
		fprintf(outfile, "\t}\n\n");
		material = material->next;
		}
	}

/* dump a list of Material_tables */

static void dump_material_tables(VDF_MATERIAL_TABLE *table)
	{
	while (table)
		{
		int i;
		fprintf(outfile, "Material_table\n\t{\n");
		if (table->n_used)  /* put out an identifier if this table is referenced */
			fprintf(outfile, "\tIdentifier { 0x%lX }\n", (long) table->id);
		if (table->application_handle)
			fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) table->application_handle);
		fprintf(outfile, "\tCount { %ld }\n", (long) table->count);
		for (i = 0; i < table->count; ++i)
			fprintf(outfile, "\tMaterial_reference { 0x%lX }\n", (long) table->material_references[i]->id);
		fprintf(outfile, "\t}\n\n");
		table = table->next;
		}
	}

/* dump a single Vertex */

static void dump_vertex(VDF_VERTEX *v)
	{
	if (v == NULL)
		return;
	fprintf(outfile, "\t\tVertex\n\t\t\t{\n");
	fprintf(outfile, "\t\t\tPoint3D { %f %f %f }\n", v->vector.x, v->vector.y, v->vector.z);
	if (vector_non_zero(&v->normal))
		fprintf(outfile, "\t\t\tNormal3D { %f %f %f }\n", v->normal.x, v->normal.y, v->normal.z);
	if (color_non_zero(&v->color))
		fprintf(outfile, "\t\t\tColor { %f %f %f }\n", v->color.red, v->color.green, v->color.blue);
	if (v->application_handle)
		fprintf(outfile, "\t\t\tApplication_handle { 0x%lX }\n", (long) v->application_handle);
	fprintf(outfile, "\t\t\t}\n");
	}

/* dump a single Facet */

static void dump_facet(VDF_FACET *f)
	{
	if (f == NULL)
		return;
	fprintf(outfile, "\t\tFacet\n\t\t\t{\n");
	if (f->n_used)  /* if we're referenced, we need to have an identifier */
		fprintf(outfile, "\t\t\tIdentifier { 0x%lX }\n", (long) f->id);
	if (f->is_doublesided)  /* default is FALSE */
		fprintf(outfile, "\t\t\t\Is_doublesided { TRUE }\n");
	if (f->is_interior)     /* default is FALSE */
		fprintf(outfile, "\t\t\t\Is_interior { TRUE }\n");
	if (f->base_facet)
		fprintf(outfile, "\t\t\tBase_facet { 0x%lX }\n", (long) f->base_facet->id);
	if (f->front_material)
		fprintf(outfile, "\t\t\tFront_material { %ld }\n", (long) f->front_material);
	if (f->back_material)
		fprintf(outfile, "\t\t\tBack_material { %ld }\n", (long) f->back_material);
	if (vector_non_zero(&f->normal))
		fprintf(outfile, "\t\t\tNormal3D { %f %f %f }\n", f->normal.x, f->normal.y, f->normal.z);
	if (f->application_handle)
		fprintf(outfile, "\t\t\tApplication_handle { 0x%lX }\n", (long) f->application_handle);
	if (f->npoints && f->points)
		{
		int i;
		fprintf(outfile, "\t\t\tVertex_data\n\t\t\t\t{\n");
		fprintf(outfile, "\t\t\t\tCount { %ld }\n", (long) f->npoints);
		for (i = 0; i < f->npoints; ++i)
			fprintf(outfile, "\t\t\t\tVertex_info { Index { %ld } }\n", (long) f->points[i].vertnum);
		fprintf(outfile, "\t\t\t\t}\n");
		}
	fprintf(outfile, "\t\t\t}\n");
	}

/* dump a list of Shapes */

static void dump_shapes(VDF_SHAPE *shape)
	{
	while (shape)
		{
		fprintf(outfile, "Shape\n\t{\n");
		if (shape->lod_size != 0.00)
			fprintf(outfile, "\tLOD_size { %f }\n", shape->lod_size);
		if (shape->replaces)
			fprintf(outfile, "\tLOD_replaces { 0x%lX }\n", (long) shape->replaces->id);
		if (shape->n_used)
			fprintf(outfile, "\tIdentifier { 0x%lX }\n", (long) shape->id);
		if (shape->name)
			fprintf(outfile, "\tName { \"%s\" }\n", shape->name);
		if (shape->is_convex)
			fprintf(outfile, "\tIs_convex { TRUE }\n");
		if (shape->application_handle)
			fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) shape->application_handle);
		if (vector_non_zero(&shape->bound_min) || vector_non_zero(&shape->bound_max))
			{
			fprintf(outfile, "\tBounding_box { %f %f %f %f %f %f }\n",
				shape->bound_min.x, shape->bound_min.y, shape->bound_min.z,
				shape->bound_max.x, shape->bound_max.y, shape->bound_max.z);
			}
		if (shape->material_table)
			fprintf(outfile, "\tUses_material_table { 0x%lX }\n", (long) shape->material_table->id);
		if (shape->vertices)
			{
			int i;
			fprintf(outfile, "\tVertex_list\n\t\t{\n");
			fprintf(outfile, "\t\tCount { %d }\n", shape->nvertices);
			for (i = 0; i < shape->nvertices; ++i)
				dump_vertex(&shape->vertices[i]);
			fprintf(outfile, "\t\t}\n");
			}
		if (shape->facets)
			{
			int i;
			fprintf(outfile, "\tFacet_list\n\t\t{\n");
			fprintf(outfile, "\t\tCount { %d }\n", shape->nfacets);
			for (i = 0; i < shape->nfacets; ++i)
				dump_facet(&shape->facets[i]);
			fprintf(outfile, "\t\t}\n");
			}
		fprintf(outfile, "\t}\n\n");
		shape = shape->next;
		}
	}

/* dump a list of Objects */

static void dump_objects(VDF_OBJECT *object)
	{
	while (object)
		{
		if (object->n_used || object->shape)
			{
			fprintf(outfile, "Object\n\t{\n");
			if (object->name)
				fprintf(outfile, "\tName { \"%s\" }\n", object->name);
			if (object->n_used)
				fprintf(outfile, "\tIdentifier { 0x%lX }\n", (long) object->id);
			if (object->shape)
				fprintf(outfile, "\tInstance_of_shape { 0x%lX }\n", (long) object->shape->id);
			if (object->scaled_by.x != 1.0 || object->scaled_by.y != 1.0 || object->scaled_by.z != 1.0)
				fprintf(outfile, "\tScaled_by { %f %f %f }\n", object->scaled_by.x, object->scaled_by.y, object->scaled_by.z);
			if (object->material_table)
				fprintf(outfile, "\tUses_material_table { 0x%lX }\n", (long) object->material_table->id);
			if (vector_non_zero(&object->location))
				fprintf(outfile, "\tLocation { %f %f %f }\n", object->location.x, object->location.y, object->location.z);
			if (object->rotx != 0.0 || object->roty != 0.0 || object->rotz != 0.0)
				fprintf(outfile, "\tRotation { %f %f %f }\n", object->rotx, object->roty, object->rotz);
			if (object->parent)
				fprintf(outfile, "\tAttached_to { 0x%lX }\n", (long) object->parent->id);
			if (object->container)
				fprintf(outfile, "\tContained_within { 0x%lX }\n", (long) object->container->id);
			if (object->is_invisible)
				fprintf(outfile, "\tIs_invisible { TRUE }\n");
			if (object->layer)
				fprintf(outfile, "\tLayer { %ld }\n", (long) object->layer);
			if (object->text)
				fprintf(outfile, "\tText { \"%s\" }\n", object->text);
			if (object->facet_behind)
				fprintf(outfile, "\tFacet_behind { 0x%lX }\n", (long) object->facet_behind->id);
			if (object->application_handle)
				fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) object->application_handle);
			fprintf(outfile, "\t}\n\n");
			}
		object = object->next;
		}
	}

static char *light_type_names[] = { "DIRECTIONAL", "POINT", "SPOT" };

/* dump a list of Lights */

static void dump_lights(VDF_LIGHT *light)
	{
	while (light)
		{
		fprintf(outfile, "Light\n\t{\n");
		if (light->name)
			fprintf(outfile, "\tName { \"%s\" }\n", light->name);
		if (light->application_handle)
			fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) light->application_handle);
		if (light->type != VDF_LIGHT_DIRECTIONAL)
			fprintf(outfile, "\tType { %s }\n", light_type_names[light->type]);
		if (light->associated_with)
			fprintf(outfile, "\tAssociated_with { 0x%lX }\n", (long) light->associated_with->id);
		if (color_non_zero(&light->color))
			fprintf(outfile, "\tColor { %f %f %f }\n", light->color.red, light->color.green, light->color.blue);
		if (light->hotspot)
			fprintf(outfile, "\tHotspot { %f }\n", light->hotspot);
		if (light->falloff)
			fprintf(outfile, "\tFalloff { %f }\n", light->falloff);
		if (!light->is_on)  /* default is TRUE */
			fprintf(outfile, "\tIs_on { FALSE }\n");
		if (light->casts_shadows)
			fprintf(outfile, "\tCasts_shadows { TRUE }\n");
		fprintf(outfile, "\t}\n\n");
		light = light->next;
		}
	}

static char *proj_type_names[] = { "PERSPECTIVE", "PARALLEL" };

/* dump a list of Cameras */

static void dump_cameras(VDF_CAMERA *camera)
	{
	while (camera)
		{
		fprintf(outfile, "Camera\n\t{\n");
		if (camera->name)
			fprintf(outfile, "\tName { \"%s\" }\n", camera->name);
		if (camera->application_handle)
			fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) camera->application_handle);
		if (camera->field_of_view)
			fprintf(outfile, "\tField_of_view { %f }\n", camera->field_of_view);
		if (camera->aspect_ratio != 1.33)
			fprintf(outfile, "\tAspect_ratio { %f }\n", camera->aspect_ratio);
		if (camera->associated_with)
			fprintf(outfile, "\tAssociated_with { 0x%lX }\n", (long) camera->associated_with->id);
		if (camera->projection_type != VDF_PROJ_PERSPECTIVE)
			fprintf(outfile, "\tProjection_type { %s }\n", proj_type_names[camera->projection_type]);
		fprintf(outfile, "\t}\n\n");
		camera = camera->next;
		}
	}

/* dump a list of Sounds */

static void dump_sounds(VDF_SOUND *sound)
	{
	while (sound)
		{
		fprintf(outfile, "Sound\n\t{\n");
		if (sound->name)
			fprintf(outfile, "\tName { \"%s\" }\n", sound->name);
		if (sound->application_handle)
			fprintf(outfile, "\tApplication_handle { 0x%lX }\n", (long) sound->application_handle);
		if (sound->associated_with)
			fprintf(outfile, "\tAssociated_with { 0x%lX }\n", (long) sound->associated_with->id);
		if (!sound->is_on)  /* default is TRUE */
			fprintf(outfile, "\tIs_on { FALSE }\n");
		if (sound->volume)
			fprintf(outfile, "\tVolume { %f }\n", sound->volume);
		if (sound->sample_name)
			fprintf(outfile, "\tSample_name { \"%s\" }\n", sound->sample_name);
		fprintf(outfile, "\t}\n\n");
		sound = sound->next;
		}
	}

/* dumps the information contained with a Created_by or Modified_by tag */

static void dump_human_data(VDF_HUMAN_DATA *h)
	{
	if (h == NULL)
		return;
	if (h->person)
		fprintf(outfile, "\t\tPerson { \"%s\" }\n", h->person);
	if (h->date || h->time)
		fprintf(outfile, "\t\tDate { %s %s }\n", h->date ? h->date : "", h->time ? h->time : "");
	}

/* dump the World_information */

static void dump_world_info(VDF_DATA *ptr)
	{
	VDF_HUMAN_DATA *p;
	if (ptr == NULL)
		return;
	if (!ptr->created && !ptr->modified && !ptr->copyright_message && !ptr->usage_restrictions && !ptr->title)
		return;
	fprintf(outfile, "World_information\n\t{\n");
	if (ptr->title)
		fprintf(outfile, "\tTitle { \"%s\" }\n", ptr->title);
	if (ptr->created)
		{
		fprintf(outfile, "\tCreated_by\n\t\t{\n");
		dump_human_data(ptr->created);
		fprintf(outfile, "\t\t}\n");
		}
	for (p = ptr->modified; p; p = p->next)
		{
		fprintf(outfile, "\tModified_by\n\t\t{\n");
		dump_human_data(p);
		fprintf(outfile, "\t\t}\n");
		}
	if (ptr->copyright_message)
		fprintf(outfile, "\tCopyright_message { \"%s\" }\n", ptr->copyright_message);
	if (ptr->usage_restrictions)
		fprintf(outfile, "\tUsage_restrictions_message { \"%s\" }\n", ptr->usage_restrictions);
	fprintf(outfile, "\t}\n\n");
	}

/* dump the World_attributes */

static void dump_world_attributes(VDF_WORLD_ATTRIBUTES *attrs)
	{
	if (attrs == NULL)
		return;
	fprintf(outfile, "World_attributes\n\t{\n");
	if (vector_non_zero(&attrs->gravity_vector))
		fprintf(outfile, "\tGravity_vector { %f %f %f }\n", attrs->gravity_vector.x, attrs->gravity_vector.y, attrs->gravity_vector.z);
	if (color_non_zero(&attrs->ambient_light))
		fprintf(outfile, "\tAmbient_light { %f %f %f }\n", attrs->ambient_light.red, attrs->ambient_light.green, attrs->ambient_light.blue); 
	fprintf(outfile, "\tHas_horizon { %s }\n", attrs->has_horizon ? "TRUE" : "FALSE");
	if (color_non_zero(&attrs->sky_color))
		fprintf(outfile, "\tSky_color { %f %f %f }\n", attrs->sky_color.red, attrs->sky_color.green, attrs->sky_color.blue); 
	if (color_non_zero(&attrs->ground_color))
		fprintf(outfile, "\tGround_color { %f %f %f }\n", attrs->ground_color.red, attrs->ground_color.green, attrs->ground_color.blue); 
	if (color_non_zero(&attrs->fog_color))
		fprintf(outfile, "\tFog_color { %f %f %f }\n", attrs->fog_color.red, attrs->fog_color.green, attrs->fog_color.blue); 
	if (attrs->scale != 1.00)
		fprintf(outfile, "\tScale { %f }\n", attrs->scale);
	fprintf(outfile, "\t}\n\n");
	}

/* dump the file by calling all the routines defined above */

static void dump_file(VDF_DATA *ptr)
	{
	fprintf(outfile, "// Saved by VDF output routines\n\n");
	dump_world_info(ptr);
	dump_palette(ptr->palette);
	dump_maps(ptr->maps);
	dump_materials(ptr->materials);
	dump_material_tables(ptr->material_tables);
	dump_shapes(ptr->shapes);
	dump_objects(ptr->objects);
	dump_lights(ptr->lights);
	dump_cameras(ptr->cameras);
	dump_sounds(ptr->sounds);
	dump_world_attributes(ptr->world_attributes);
	fprintf(outfile, "// End of file\n");
	}

/**** main file-writing routine ****/

int vdf_writefile(char *filename, VDF_DATA *ptr)
	{
	strcpy(emsg, "data structure pointer is NULL");
	if (ptr == NULL)
		return -2;
	sprintf(emsg, "could not open '%s'", filename);
	outfile = fopen(filename, "w");
	if (outfile == NULL)
		return -1;
	strcpy(emsg, "No read error");
	if (error_setup())
		return -3;
	dump_file(ptr);
	return 0;
	}

